#!/usr/bin/env python
import numpy as np
from datetime import datetime, timedelta
from scipy.interpolate import CubicSpline, interp1d
import matplotlib.pyplot as plt
import re
import os
#from datetime import datetime

__authors__ = ["Ming Chen"]
__license__ = "Public Domain"
__maintainer__ = "Ming Chen"
__email__ = "ming.chen@noaa.gov"
__date__  = "Mar 24, 2023"

'''
#####################################################################
# 
# leopod_sp3.py
#
# Purpose:
# This python script has the functions to  combine the  orbit and clock 
# into a SP3 file of a specific range of epoches 
#
# History:  
#       Created by: Ming Chen 2026/02/05
#                   ming.chen@noaa.gov
# 
# Usage: python leopod_sp3.py orbit_file clock_file sp3_file 
# --start-time yyyymnddThhmmss --end-time yyyymnddThhmmss
#
#
#####################################################################

'''

# ------------------------------------------------------------
# Parse the clock file (RINEX CLK-like format)
# ------------------------------------------------------------

def parse_clock_file(clock_file):
    """
    Reads a clock file and returns a dictionary:
        clock[(sat, datetime)] = clock_value
    """
    clock_dict = {}

    with open(clock_file, "r") as f:
        for line in f:
            if not line.startswith("AR"):
                continue

            parts = line.split()

            # Satellite ID (e.g., G29)
            sat = parts[1].strip()
            if sat not in clock_dict:
               clock_dict[sat] = {"times": [], "CLK": []}

            # Date-time fields
            year  = int(parts[2])
            month = int(parts[3])
            day   = int(parts[4])
            hour  = int(parts[5])
            minute = int(parts[6])
            sec    = float(parts[7])

            # Build datetime object
            current_time = datetime(
                year, month, day, hour, minute,
                int(sec),
                int((sec % 1) * 1e6)
            )

            #current_time = datetime(int(year), int(month), int(day),
            #                        int(hour), int(minute), int(float(second)))
            # Clock offset (seconds)
            clk = float(parts[9])*1e6
            clock_dict[sat]["times"].append(current_time )
            clock_dict[sat]["CLK"].append(clk)

    return clock_dict



# ------------------------ Read SP3 safely ------------------------

def read_sp3_with_regex(file_path, clock_unit='us', t_start:datetime = None, t_end:datetime = None, check_clk=False):
    """
    Read SP3 including positions, velocities, and clock using regex.
    Returns data dictionary and header lines.
    """
    data = {}
    header_lines = []

    # Regular expressions
    epoch_pattern = re.compile(r'^\*\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)\s+([\d.]+)')
    pos_pattern   = re.compile(r'^P(\w{3})\s*([-\d.Ee+]+)\s*([-\d.Ee+]+)\s*([-\d.Ee+]+)\s*([-\d.Ee+]+)')
    vel_pattern   = re.compile(r'^V(\w{3})\s*([-\d.Ee+]+)\s*([-\d.Ee+]+)\s*([-\d.Ee+]+)')

    current_time = None
    last_sat = None

    with open(file_path, 'r') as f:
        lines = f.readlines()

    # Extract header
    for line in lines:
        if line.startswith('*'):
            break
        header_lines.append(line.rstrip('\n'))
    
    if t_start is None: t_start = datetime(1900, 1, 1) 
    if t_end is None:   t_end = datetime(2900, 1, 1)
 
    # Parse data
    for line in lines:
        line = line.rstrip('\n')
        if m := epoch_pattern.match(line):
            year, month, day, hour, minute, second = m.groups()
            current_time = datetime(int(year), int(month), int(day),
                                    int(hour), int(minute), int(float(second)),  int((float(second) % 1) * 1e6))

        elif m := pos_pattern.match(line):
            if current_time < t_start or  current_time > t_end: continue
            sat, x, y, z, clk = m.groups()
            x, y, z, clk = map(float, (x, y, z, clk))
            if clk > 99999 and check_clk: continue
            if sat not in data:
                data[sat] = {"times": [], "X": [], "Y": [], "Z": [], "VX": [], "VY": [], "VZ": [], "CLK": []}
           
            data[sat]["times"].append(current_time)
            data[sat]["X"].append(x)
            data[sat]["Y"].append(y)
            data[sat]["Z"].append(z)
            data[sat]["CLK"].append(clk)
            last_sat = sat
        elif m := vel_pattern.match(line):
            if current_time < t_start or  current_time > t_end: continue
            sat, vx, vy, vz = m.groups()
            vx, vy, vz = map(float, (vx, vy, vz))
            if sat in data:
                data[sat]["VX"].append(vx)
                data[sat]["VY"].append(vy)
                data[sat]["VZ"].append(vz)
    return data, header_lines



# ------------------------ Compute velocities if missing or at high resolution ------------------------
def compute_velocity_from_position(sat_data):
    """
    Compute VX, VY, VZ from positions and times using finite differences.
    """
    times_sec = np.array([(t - sat_data["times"][0]).total_seconds() for t in sat_data["times"]])
    scale = 1.0e4
    X = np.array(sat_data["X"])
    Y = np.array(sat_data["Y"])
    Z = np.array(sat_data["Z"])
    n = len(times_sec)
    VX = np.zeros(n)
    VY = np.zeros(n)
    VZ = np.zeros(n)

    # Central differences
    for i in range(1, n-1):
        dt = times_sec[i+1] - times_sec[i-1]
        VX[i] = (X[i+1] - X[i-1]) / dt * scale
        VY[i] = (Y[i+1] - Y[i-1]) / dt * scale
        VZ[i] = (Z[i+1] - Z[i-1]) / dt * scale

    # Forward difference for first point
    VX[0] = (X[1] - X[0]) / (times_sec[1] - times_sec[0]) * scale
    VY[0] = (Y[1] - Y[0]) / (times_sec[1] - times_sec[0]) * scale
    VZ[0] = (Z[1] - Z[0]) / (times_sec[1] - times_sec[0]) * scale

    # Backward difference for last point
    VX[-1] = (X[-1] - X[-2]) / (times_sec[-1] - times_sec[-2]) * scale
    VY[-1] = (Y[-1] - Y[-2]) / (times_sec[-1] - times_sec[-2]) * scale
    VZ[-1] = (Z[-1] - Z[-2]) / (times_sec[-1] - times_sec[-2]) * scale

    return VX.tolist(), VY.tolist(), VZ.tolist()



# ------------------------ Interpolation ------------------------
def interpolate_sat_data_cubic(sat_data, new_interval=30.0, new_times:list = [], ds_keys:list=[]):
    """
    Interpolates a single satellite:
    - X/Y/Z, VX/VY/VZ: cubic spline
    - CLK: linear
    """

    if len(ds_keys) < 1:
        ds_keys = sat_data.keys()
    if 'times' not in ds_keys:
        print(f"'times' not in sat_data ...")
        return
    
    tref =  datetime(2000,1,1)

    times_sec = np.array([(t - tref).total_seconds() for t in sat_data["times"]])
    
    if len(new_times)<1 :
       new_times_sec = np.arange(0, times_sec[-1] + new_interval, new_interval)
    else:
       new_times_sec = np.array([(t - tref).total_seconds() for t in new_times])

    ds_keys.remove('times')
    interp_data = {}

   # for comp in ["X","Y","Z","VX","VY","VZ"]:
    for comp in ds_keys:
        if comp in sat_data and len(sat_data[comp]) > 0:
            cs = CubicSpline(times_sec, sat_data[comp])
            interp_data[comp] = cs(new_times_sec)
            interp_data["times"] = [tref + timedelta(seconds=t) for t in new_times_sec]
    #if 'CLK' in ds_keys:
    #  cs = CubicSpline(times_sec,sat_data['CLK'])
    #  csclk =  cs(new_times_sec)
    #  interp_data["CLK"] = csclk
    #  interp_data["times"] = [tref+ timedelta(seconds=t) for t in new_times_sec]
    return interp_data

# ------------------------ Interpolation ------------------------
def interpolate_sat_data_linear(sat_data, new_interval=30.0, new_times:list = [], ds_keys:list=[]):
    """
    Interpolates a single satellite:
    - X/Y/Z, VX/VY/VZ: cubic spline
    - CLK: linear
    """

    if len(ds_keys) < 1:
        ds_keys = sat_data.keys()
    if 'times' not in ds_keys:
        print(f"'times' not in sat_data ...")
        return
    
    tref =  datetime(2000,1,1)

    times_sec = np.array([(t - tref).total_seconds() for t in sat_data["times"]])
    
    if len(new_times)<1 :
       new_times_sec = np.arange(0, times_sec[-1] + new_interval, new_interval)
    else:
       new_times_sec = np.array([(t - tref).total_seconds() for t in new_times])

    ds_keys.remove('times')
    interp_data = {}

   # for comp in ["X","Y","Z","VX","VY","VZ"]:
    for comp in ds_keys:
        if comp in sat_data and len(sat_data[comp]) > 0:
            cs = interp1d(times_sec, sat_data[comp], kind="linear", fill_value="extrapolate")
            interp_data[comp] = cs(new_times_sec)
            interp_data["times"] = [tref + timedelta(seconds=t) for t in new_times_sec]


    return interp_data


def interpolate_all_satellites(data_dict, new_interval=30.0):
    return {sat: interpolate_satellite_data(sat_data, new_interval)
            for sat, sat_data in data_dict.items()}

# ------------------------ Satellite sorting ------------------------
def sat_sort_key(sat):
    """Sort satellites G E C then by number"""
    system_order = {'G':0,'E':1,'C':2}
    sys = sat[0]
    num = int(sat[1:])
    return (system_order.get(sys,99), num)


def write_sp3_highres_with_velocity(output_file, interpolated_data, header_lines, resolution=30.0):
    """
    Write high-resolution SP3 including positions, velocities, clock
    - Position line is followed by velocity line
    - Velocity line gets default CLK value 999999.999999
    """
    default_clk_for_velocity = 999999.999999
    hres   = re.compile(r'(##\s*\d+\s+[-\d.Ee+]+)\s+([-\d.Ee+]+)(.+)')

    with open(output_file, 'w') as f:
        # Header
        for line in header_lines:
            resline = line.rstrip('\n')
            if m := hres.match(resline):
               h1,h2,h3 = m.groups()
               line = h1 + f"{resolution:15.8f}" + h3
            f.write(line + '\n')

        # Sort satellites
        if len(interpolated_data.keys()) > 1:
          all_sats = sorted(interpolated_data.keys(), key=sat_sort_key)
        else:
          all_sats =  list(interpolated_data.keys() )
        num_epochs = len(interpolated_data[all_sats[0]]["times"])

        for i in range(num_epochs):
            t = interpolated_data[all_sats[0]]["times"][i]
            # Epoch line
            f.write("*  {:4d} {:2d} {:2d} {:2d} {:2d} {:11.8f}\n".format(
                t.year, t.month, t.day, t.hour, t.minute, t.second + t.microsecond*1e-6
            ))
            for sat in all_sats:
                # Position line
                f.write("P{:<3s}{:14.6f}{:14.6f}{:14.6f}{:14.6f}\n".format(
                    sat,
                    interpolated_data[sat]["X"][i],
                    interpolated_data[sat]["Y"][i],
                    interpolated_data[sat]["Z"][i],
                    interpolated_data[sat]["CLK"][i]
                ))

                # Velocity line only if exists
                vx_array = interpolated_data[sat].get("VX")
                if vx_array is not None and len(vx_array) > 0:
                    f.write("V{:<3s}{:14.6f}{:14.6f}{:14.6f}{:14.6f}\n".format(
                        sat,
                        vx_array[i],
                        interpolated_data[sat]["VY"][i],
                        interpolated_data[sat]["VZ"][i],
                        default_clk_for_velocity
                    ))


# ------------------------ Plotting ------------------------
def plot_satellite_data(original_data, interpolated_data, satellites_to_plot):
    for sat in satellites_to_plot:
        if sat not in original_data:
            continue
        orig = original_data[sat]
        interp = interpolated_data[sat]

        # Positions + Clock
        fig, axs = plt.subplots(4,1, figsize=(12,10), sharex=True)
        for i, comp in enumerate(["X","Y","Z","CLK"]):
            axs[i].plot(orig["times"], orig[comp], 'o-', label="Original")
            axs[i].plot(interp["times"], interp[comp], '-', label="Interpolated")
            axs[i].set_ylabel(comp)
            axs[i].grid(True)
            axs[i].legend()
        axs[3].set_xlabel("Time")
        plt.suptitle(f"Satellite {sat} Position + Clock")
        plt.tight_layout()
        plt.show()

        # Velocities
        if len(orig["VX"]) > 0:
            fig, axs = plt.subplots(3,1, figsize=(12,8), sharex=True)
            for i, comp in enumerate(["VX","VY","VZ"]):
                axs[i].plot(orig["times"], orig[comp], 'o-', label="Original")
                axs[i].plot(interp["times"], interp[comp], '-', label="Interpolated")
                axs[i].set_ylabel(comp)
                axs[i].grid(True)
                axs[i].legend()
            axs[2].set_xlabel("Time")
            plt.suptitle(f"Satellite {sat} Velocities")
            plt.tight_layout()
            plt.show()


def normalize_to_datetime(t):
    """Ensure all times are Python datetime"""
    if isinstance(t, datetime):
        return t
    else:
        return t.astype('datetime64[ns]').astype(datetime)


def filter_dict_by_time(data, start_time, end_time):

    times = [normalize_to_datetime(t) for t in data["times"]]
 
    indices = [
        i for i, t in enumerate(times)
        if start_time <= t <= end_time
    ]

    filtered = {
        key: [values[i] for i in indices]
        for key, values in data.items()
    }

    return filtered   


def interpolate_leopod_finclk(fin_file, clk_file, fin_output, 
                            t_start:datetime, t_end:datetime, sliding=60, algo='linear',leostatus=''):

   if not os.path.exists(fin_file) or not os.path.exists(clk_file):
      exit
   
   print(f"FIN Input: {fin_file}")
   print(f"CLK Input: {clk_file}")
   print(f"FIN_Output: {fin_output}")
   #fname = os.path.basename(fin_file).split('_')
   #tend = fname[-2]
   #t_end = datetime.strptime(tend, "%Y%m%dT%H%M%S")
   #t_start = t_end - timedelta(minutes=sliding)

   # Load clock values
   clk_data = parse_clock_file(clk_file)
   sats = list(clk_data.keys()) ; sat = sats[0]
   clkpod = clk_data[sat]
   leosats = {'L23': 'C2E1', 'L24': 'C2E2','L25': 'C2E3',\
               'L26': 'C2E4', 'L27': 'C2E5','L28': 'C2E6'}

   leosats = {'C2E1':'L23', 'C2E2':'L24', 'C2E3':'L25',\
              'C2E4':'L26', 'C2E5':'L27', 'C2E6':'L28'}

   sat = leosats[sat]
   

   # Read file
   #t_start = kinpod['times'][0] ; t_end =  kinpod['times'][-1]
   #fin_data, fin_header_lines =  read_sp3_with_regex(fin_file, t_start = t_start, t_end=t_end)
   fin_data, fin_header_lines =  read_sp3_with_regex(fin_file)
   if leostatus:
      leostatus = f"/* LEOSTATUS {leostatus} \n"
      fin_header_lines[-1] = leostatus 

   finpod = fin_data[sat] 
   if len(finpod['VX']) < 1:
      finpod['VX'], finpod['VY'], finpod['VZ'] = compute_velocity_from_position(finpod)
  
   new_times = finpod['times']
   #interp_data = interpolate_sat_data(kinpod, new_interval=resolution)
   if algo == 'cubic':
      interp_data = interpolate_sat_data_cubic(clkpod, new_times=new_times, ds_keys=['times','CLK'] )
   else:
      interp_data = interpolate_sat_data_linear(clkpod, new_times=new_times, ds_keys=['times','CLK'] )

   finpod['CLK'] = interp_data['CLK']
   #print(finpod['times'])
   
   
   finpod = filter_dict_by_time(finpod, t_start, t_end)
   fin_data[sat] = finpod
   
   write_sp3_highres_with_velocity(fin_output, fin_data, fin_header_lines)
   


# ------------------------ MAIN ------------------------
if __name__ == "__main__":
   import argparse

   parser = argparse.ArgumentParser(description='merge orbit and clock into SP3 files  ')
   parser.add_argument("-v","--verbose", action='store_true',  help="print info for debug")
   parser.add_argument("fin_input", type=str,  help="file with only position  ")
   parser.add_argument("clk_input", type=str,  help="file with leo clock")
   parser.add_argument("fin_output",type=str,  help="Output ORB SP3 file at higher resolution")
   parser.add_argument("-start", "--start-time", type=str,  help="%Y%m%dT%H%M%S")
   parser.add_argument("-end", "--end-time", type=str,  help="%Y%m%dT%H%M%S")
   parser.add_argument("-status", "--leo-status", type=str, help="LEO Bernese Solution Status")
   parser.add_argument("-sliding", "--sliding-minutes", type=int, default=30, help="sliding minutes")

   args = parser.parse_args()
    
   fin_input = args.fin_input  ; clk_input = args.clk_input
   fin_output = args.fin_output
   sliding = args.sliding_minutes
   t_start = args.start_time
   t_end = args.end_time
   tstart = datetime.strptime(t_start, "%Y%m%dT%H%M%S")
   tend =  datetime.strptime(t_end, "%Y%m%dT%H%M%S")
   leostatus = args.leo_status 
 
   interpolate_leopod_finclk(fin_input, clk_input, fin_output, tstart, tend, sliding=sliding, algo='linear',leostatus=leostatus)



#!/usr/bin/env python
import os, sys, glob, subprocess, shutil
from datetime import datetime, timedelta
import re, string
import numpy as np
from satellite_info import env_paths, sat_ant_info

from ucar2umd_att import UCAR2UMD_ATT
from utility import str2date, leofile_time_range
from ucar_gnssro_data import local_ucar_filelist

__authors__ = ["Ming Chen"]
__license__ = "Public Domain"
__maintainer__ = "Ming Chen"
__email__ = "ming.chen@noaa.gov"
__date__  = "April 24, 2026"

'''
#####################################################################
# 
# ucar2umd_att_rt.py
#
# Purpose:
# This Python driver script merges UCAR sub-daily LEO attitude data files
# into a unified file over a user-defined epoch range to support
# real-time POD processing with the BERNESE software package.
##
# History:  
#       Created by: Ming Chen 2026/04/24
#                   ming.chen@noaa.gov
# 
# Usage: python  ucar2umd_att_rt.py  --verbose mission satidx sdate edate
#
# Args:  mission[cosmic2, spire, cosmic1] satidx 
#        start-date[yyyyddd|yyyymndd] [end-date] --window-hours [1-12]
#
#####################################################################

'''
def umdrepo_att_filelist(mission, satidx,  sdate, **kwargs):      
    # Check if required environment variables are set

    umd_repo = kwargs.get('umd_repo','')
    satinfo = sat_ant_info(mission, satidx)
    if satinfo == None:return None
    satsvn =  satinfo['satsvn'] ; satname = satinfo['satname'] 
    year = sdate.strftime('%Y') 
    jday = sdate.strftime('%j')
    #leoAtt_path =  os.path.join(umd_repo, mission,'level1a', 'leoAtt', year, jday)
    leoAtt_path =  os.path.join(umd_repo, year, jday)
    fnames = 'leoAtt_{}.{}.{}.*'.format(year, jday, satsvn)
    filelist = glob.glob(leoAtt_path+'/'+fnames+'_txt', recursive=False)       
    return filelist 

def umdatt_filepath(mission, pdate):
    umddir  = os.getenv('umd_data')
    year = pdate.strftime('%Y') ; jday = pdate.strftime('%j')
    yday = f'{year}.{jday}' 
    umdLeoAtt_Path =  os.path.join(umddir, mission, 'leoAtt', yday)
    
    return umdLeoAtt_Path

def umdatt_filename(mission, satidx, tstart, tend):
    satinfo = sat_ant_info(mission, satidx)
    if satinfo == None:return None
    satsvn =  satinfo['satsvn'] ; satname = satinfo['satname'] 
    t1 = tstart.strftime('%Y%m%dT%H%M%S')
    t2 = tend.strftime('%Y%m%dT%H%M%S')
    outfile = f"{mission}_{t1}_{t2}_{satsvn}.att"
    return outfile


def ucar2umd_att_session(mission, satidx, tstart, tend, **kwargs):      
    
    alpha=list(string.ascii_uppercase)
    verbose = kwargs.get('verbose') 
    umd_repo = kwargs.get('umd_repo','')
    grace_hours = kwargs.get('grace_hours',3)
    if (verbose == None): 
      verbose = os.environ['_VERBOSE_'].lower() in ('true', '1','yes')
    overwrite = kwargs.get('overwrite') 
    if (overwrite == None): 
      overwrite = os.environ['_OVERWRITE_'].lower() in ('true', '1','yes')
   
    satinfo = sat_ant_info(mission, satidx)
    if satinfo == None: return
    satsvn =  satinfo['satsvn'] ; satname = satinfo['satname'] 
    

    dt = timedelta(hours=grace_hours)
    ndays = (tend - tstart + dt + dt).days
    sdate = (tstart - dt).strftime('%Y%m%d') ; sdate = datetime.strptime(sdate,'%Y%m%d')
    edate = (tend + dt).strftime('%Y%m%d') ; edate = datetime.strptime(edate,'%Y%m%d')
    print(sdate, edate)
    flists=[]
    while sdate <= edate:
       year = sdate.strftime('%Y') 
       jday = sdate.strftime('%j')
       yday=f'{year}.{jday}'
       datestr = sdate.strftime('%Y%m%d')

       if umd_repo.strip():
          filelist =  umdrepo_att_filelist(mission, satidx,  sdate, umd_repo = umd_repo)
       else:
          filelist =  local_ucar_filelist(mission, 'leoAtt', datestr, svn=satsvn)
       flists.extend(filelist)
       sdate = sdate + timedelta(days=1)  
    filelist = sorted(flists)

    if len(filelist) < 1: return None
     
    umd = UCAR2UMD_ATT(mission=mission)
    n = 0
    for f in filelist:
        rr=leofile_time_range(f)
        r1 = datetime.strptime(rr[0],'%Y%m%dT%H:%M:%S.%f')
        r2 = datetime.strptime(rr[1],'%Y%m%dT%H:%M:%S.%f')
        #r1 = rr[0] ; r2 = rr[1]
        if  rr[0] == None : continue
        if  r1 > tend or r2 < tstart :  continue
        #print(f)
        _, basename = os.path.split(os.path.abspath(f))
        umd.add_rinex(f, verbose=verbose)
        n = n + 1

    umdLeoAtt =  umdatt_filepath(mission, tend)
    outfile = umdatt_filename(mission, satidx, tstart, tend)
    if n != 0:  
      os.makedirs(umdLeoAtt, exist_ok=True)
      fileout = os.path.join(umdLeoAtt,  outfile) 
      if os.path.exists(fileout): 
         print('{} already existed, rewrite'.format(fileout))
      t1 = tstart.strftime('%Y%m%dT%H%M%S')
      t2 = tend.strftime('%Y%m%dT%H%M%S')
      print('time range {} {} , valid ucars files {}'.format(t1, t2, n))
 
      leoprn = 923 + satidx - 1 ; leotag = 'CHACC{:3d}'.format(leoprn)
      #leoprn = 923 ; leotag = 'CHACC{:3d}'.format(leoprn)
      #umd.write_rinex(outfile=fileout, verbose = True )  
      umd.write_rinex(outfile=fileout, satellite=leotag, verbose = True )

    else : 
      print( "no LeoAtt files found...")
      msg = '{} {} leoAtt_*{}  {}'.format(umd_repo, mission, satsvn,  datestr)
      print( msg) 
      return False

def ucar2umd_att_rt(mission, tstart, tend, satidx=-1,  **kwargs):      
    umd_repo = kwargs.get('umd_repo','')
    leoinfo = sat_ant_info(mission)
    nsat = leoinfo["nsats"]
    leosats = [i+1 for i in range(nsat)]
    if satidx > 0: leosats = [satidx]   

  
    for satidx in leosats:
        ucar2umd_att_session(mission, satidx, tstart, tend, umd_repo=umd_repo)
            

# ----------------------------------------------------------------------
# Argument parsing
# ----------------------------------------------------------------------
def parse_args():
    import argparse
    parser = argparse.ArgumentParser(
        description="Main Program arguments"
    )
   
    parser.add_argument("-t","--datetime", help = "%Y-%m-%dT%H:%M:%S")
    parser.add_argument("--window-minutes", type=float, default=60.0, help="Override minutes from YAML")
    parser.add_argument("--grace-minutes", type=float, default=0.0, help="Override grace from YAML")
    parser.add_argument("--workers", type=int, default=4)
    parser.add_argument("--dry-run", action="store_true")
    parser.add_argument("mission", type=str,  help="cosmic1 or cosmic2")
    parser.add_argument("satidx", type=int,  help="cosmic2 satidx 1-6")
    parser.add_argument("--umd-repo",  type=str, default='', help="path to umd local ucar data repository"  )
    parser.add_argument("-ucar","--ucar-data-home",type=str,  help="Local UCAR data home path "  )
    parser.add_argument("-umd","--umd-data-home",  type=str,  help="UMD data home path "  )
    parser.add_argument("-util","--utility-home",  type=str,  help="POD utility home path "  )
    parser.add_argument("-v","--verbose", action='store_true',  help="print info for debug")


    return parser.parse_args()

    
if __name__ == "__main__":
    
    # example
    # python ucar2umd_att_rt.py cosmic2 2  -t "2026-04-05T03:00:00" --umd-repo '/data2/xinjiaz/data_GPSRO/UCAR'
   
    args = parse_args()
    ddt = datetime.strptime(args.datetime, "%Y-%m-%dT%H:%M:%S")

    # Time window
    window_minutes = args.window_minutes
    grace_minutes = args.grace_minutes 
    umd_repo = args.umd_repo

    total_minutes = window_minutes + grace_minutes
    if total_minutes < 0:
       t_start = ddt + timedelta(minutes=total_minutes)
       t_end   = ddt
    else:
       t_start = ddt ; t_end   = ddt + timedelta(minutes=total_minutes)

    tstart = t_start.strftime("%Y%m%dT%H%M%S")
    tend = t_end.strftime("%Y%m%dT%H%M%S")
    mission = args.mission ; satidx = args.satidx 
    
    verbose = args.verbose

    print(satidx, tstart, tend)
    ucar2umd_att_rt(mission, satidx, t_start,t_end, verbose=verbose, umd_repo =umd_repo)





#!/usr/bin/env python
import os, sys, glob, subprocess, shutil
from datetime import datetime, timedelta
import re, string
import numpy as np
from satellite_info import env_paths, sat_ant_info
from rt_file_reproc import check_missed_leoorb
import yaml


def update_yaml_file(template_file, output_file, satant_list, replace=False):
    """
    Update 'satant' in YAML as a list of strings.

    Parameters
    ----------
    template_file : str
        Input YAML template
    output_file : str
        Output YAML file
    satant_list : list[str]
        Example:
            ["05_02=20260510T02:45:00",
             "05_02=20260510T03:15:00"]
    replace : bool
        True  -> overwrite existing satant
        False -> append to existing satant
    """

    # Load YAML
    with open(template_file, "r") as f:
        config = yaml.safe_load(f) or {}

    # Normalize existing satant ? list[str]
    existing = []
    if "satant" in config:
        if isinstance(config["satant"], dict):
            existing = [f"{k}={v}" for k, v in config["satant"].items()]
        elif isinstance(config["satant"], list):
            existing = list(config["satant"])

    # Update
    config["satant"] = satant_list if replace else existing + satant_list

    # Write YAML
    with open(output_file, "w") as f:
        yaml.dump(config, f, sort_keys=False)

    return config


# ----------------------------------------------------------------------
# Argument parsing
# ----------------------------------------------------------------------
def parse_args():
    import argparse
    parser = argparse.ArgumentParser(
        description="Main Program arguments"
    )
   
    parser.add_argument("-t","--datetime", help = "%Y-%m-%dT%H:%M:%S")
    parser.add_argument("-v","--verbose", action='store_true',  help="print info for debug")

    parser.add_argument("-o", "--yaml-output", type=str, default='updated.yaml', help="updated config yaml file")
    parser.add_argument("-i", "--yaml-input", type=str, default='rt.yaml.template', help="input config yaml file")

    return parser.parse_args()

    
if __name__ == "__main__":

    args = parse_args()
    yaml_in =   args.yaml_input
    yaml_out =  args.yaml_output
    # Load YAML
    with open(yaml_in, "r") as f:
        config = yaml.safe_load(f) or {}

    # Time window
    window_minutes = config['reproc_segment_minutes']
    total_minutes = window_minutes 
    ddt = datetime.strptime(args.datetime, "%Y-%m-%dT%H:%M:%S")
    if total_minutes < 0:
       t_start = ddt + timedelta(minutes=total_minutes)
       t_end   = ddt
    else:
       t_start = ddt ; t_end   = ddt + timedelta(minutes=total_minutes)

    tstart = t_start.strftime("%Y%m%dT%H%M%S")
    tend = t_end.strftime("%Y%m%dT%H%M%S")
    mission = config['leo']['mission']
    satidx = config['leo']['satidx'] ; antidx = config['leo']['antidx']
    umd_repo = config['leo']['rnx_repo_root']
    print(tstart, tend)

    verbose = args.verbose
    #print(mission, verbose)
    satant_list = check_missed_leoorb(mission, t_start, t_end, satidx=satidx, antidx=antidx, umd_repo = umd_repo )
    if not verbose:
      update_yaml_file( yaml_in, yaml_out, satant_list,replace=True)
    else:
      print(satant_list)




import argparse
import os, glob
import subprocess
from datetime import datetime, timedelta
import re
from gpsdate import mkdate, gpsdate
from nasa_gnss_data import local_nasa_gnss
from aiub_gnss_data import local_aiub_gnss
import shutil
from satellite_info import  sat_ant_info
from pathlib import Path

__authors__ = ["Ming Chen"]
__license__ = "Public Domain"
__maintainer__ = "Ming Chen"
__email__ = "ming.chen@noaa.gov"
__date__  = "April 22, 2026"

'''
#####################################################################
# 
# generate_bern_sessions.py
#
# Purpose:
# This python script is used to handle the BERNESE sessions.
##
# History:  
#       Created by: Ming Chen 2026/04/22
#                   ming.chen@noaa.gov
# 
# Usage: python .py start_datetime end_datetime
# the datetime in %Y-%m-%dT%H:%M:%S format
#
#
#####################################################################

'''
def bernese_datapool_subdir(mission, satidx,codcom='RTM'):
    leoinfo = sat_ant_info(mission, satidx=satidx)
    satname, satsvn= leoinfo['satname'], leoinfo['satsvn']
    subdir = f'{codcom}{satname}'
    return subdir
    

def remove_link_if_exists(link_path):
    """
    Checks if a symbolic link exists at the given path and removes it if it does.

    Args:
        link_path: The path to the symbolic link.
    """
    status = True
    if os.path.lexists(link_path):
       os.unlink(link_path)
       print(f"Symbolic link '{link_path}' removed.")

    return(status)

def bernese_path_exists():
   # bernese
   BERN_U,BERN_P,BERN_D = os.getenv("U"),os.getenv("P"),os.getenv("D")
   bern_path = os.path.isdir(BERN_U) and \
            os.path.isdir(BERN_P) and os.path.isdir(BERN_D)
   if not bern_path:
      print('BERN DataPool path, or Campaign path,not exist ...')
      return False
   else:
     return True

def bernese_gnss_inputs(gnss_prod, utc_time, data_source='nasa', codcom = 'RTM', **kwargs):
    
    gnss_latency = kwargs.get('gnss_latency', 3) 
    subdir = kwargs.get('datapool_subdir', codcom) 

    if not bernese_path_exists: return False
    pdate = utc_time.strftime('%Y%m%d')
    pdate = gpsdate(pdate)
    year, doy, ymd = pdate.year, pdate.doy, pdate.ymd
    gweek, gday = pdate.gweek, pdate.gday 
 
 
    if not bernese_path_exists: return False

    datapool = os.getenv("D")
    plink = codcom.upper() 
    gnss_flink = {
             'erp': f"{plink}{gweek}{gday}.ERP", 
             'orb': f"{plink}{gweek}{gday}.SP3", 
             'clk': f"{plink}_{year}{doy}.CLK",  
             'bia': f"{plink}_{year}{doy}.BIA"
            }
    fzip = ['.Z', '.gz']

    # gnss_prod = 'COD0OPSULT'
    gnss_time = utc_time - timedelta(hours=gnss_latency)
    ymd = gnss_time.strftime('%Y%m%d')
    h =  int(gnss_time.strftime('%H'))
    h = (h // 6) * 6
    hh = f'{h:02d}'
    if data_source.upper() == 'AIUB':
       gnss_flist = local_aiub_gnss(gnss_prod, ymd, hh=hh)
    else: 
       gnss_flist = local_nasa_gnss(gnss_prod, ymd, hh=hh)
    gnsspool_path = f"{datapool}/{subdir}/{year}{doy}"
    print(gnsspool_path)
    os.makedirs(gnsspool_path, exist_ok = True) 
    #print(gnss_flist)
    for k in gnss_flist:
       fn = gnss_flist[k] 
       fvalid = os.path.isfile(fn) and os.path.getsize(fn) > 0
       if not fvalid:
          print(f"{fn} not exists ....")
          return False

       fzip = ['.Z', '.gz']
       fsuffix = Path(fn).suffix
       fl = f'{gnsspool_path}/{gnss_flink[k]}'
       if fsuffix in fzip:
          fl = f'{fl}{fsuffix}'
   
       print(f'make link: {fn} {fl}')
       if remove_link_if_exists(fl):
             os.symlink(fn, fl)
    return True


def bernese_leornx_inputs(leo_conf:dict,**kwargs):
    if not bernese_path_exists: return False
    leo = leo_conf
    if 'mission' not in leo: return False
    
    mission = leo['mission']
    leoinfo = sat_ant_info(mission)
    if leoinfo is None:
       print(f"invalid mission {mission} ...")
       return False
    nsat = leoinfo['nsats'] ; nant =  leoinfo["nant"]
    leosats = [i+1 for i in range(nsat)]
    leoants = [i+1 for i in range(nant)]
    if leo['satidx'] > 0: leosats = [leo['satidx']]   
    if leo['antidx'] > 0: leoants = [leo['antidx']]   
    #session_tag = '0' # '0' '[A-Z]'
    session_tag = leo['session'].upper()
    mission_tag = leo['mission'].upper()
    tstart =  leo['tstart'].strftime("%Y%m%dT%H%M%S")
    tend =    leo['tend'].strftime("%Y%m%dT%H%M%S")
    mydate = leo['tend']
    year = mydate.strftime('%Y') ; doy = mydate.strftime('%j')
    yy =  mydate.strftime('%y')  
    HM = mydate.strftime('%H%M') 

    datapool = os.getenv("D"); BERN_P = os.getenv("P")
    umd_data_path  = os.getenv("umd_data") 
    leo_data_path  = os.path.join(umd_data_path,leo['mission'])
    leo_att_path = f'{leo_data_path}/leoAtt/{year}.{doy}'
    leo_rnx_path = f'{leo_data_path}/podRnx/{year}.{doy}'
    subdir = kwargs.get('datapool_subdir', mission_tag) 
    rnxpool_path  = f'{datapool}/{subdir}/{year}{doy}'
    os.makedirs(rnxpool_path, exist_ok = True) 
  
    for satidx in leosats:
       for antidx in leoants:
          leoinfo = sat_ant_info(mission, satidx=satidx, antidx=antidx)
          satname, satsvn, antsvn = leoinfo['satname'], leoinfo['satsvn'],leoinfo['antsvn'] 
 
          bernid = leoinfo['bernid']
          campaign = bernid.upper()+str(year)[2:4]+str(doy)

          # link leo files     
          frnx_file = f"{mission}_{tstart}_{tend}_{satsvn}_{antsvn}.rnx"
          frnx_link = f'{satname}{doy}{session_tag}.{year[2:]}O_{HM}_A{antsvn}'
          rnx_file0 = os.path.join(leo_rnx_path, frnx_file)
          rnx_file = os.path.join(leo_rnx_path, frnx_link)
          
          flink = os.path.join(rnxpool_path, frnx_link)
          remove_link_if_exists(flink)
          if os.path.isfile(rnx_file0):
              #shutil.move(rnx_file0,rnx_file)
              print(f'make link: {rnx_file0} {flink}')
              os.symlink(rnx_file0, flink)
              return True
          else:
              print(f'make link error: {rnx_file0} not exist ..')
              return False

def bernese_leoatt_inputs(leo_conf:dict,**kwargs):
    if not bernese_path_exists: return False
    leo = leo_conf
    if 'mission' not in leo: return False
    
    mission = leo['mission']
    leoinfo = sat_ant_info(mission)
    if leoinfo is None:
       print(f"invalid mission {mission} ...")
       return False
    nsat = leoinfo['nsats'] ; nant = leoinfo['nant']
    leosats = [i+1 for i in range(nsat)]
    leoants = [i+1 for i in range(nant)]
    if leo['satidx'] > 0: leosats = [leo['satidx']]   
    if leo['antidx'] > 0: leoants = [leo['antidx']]   
    #session_tag = '0' # '0' '[A-Z]'
    session_tag = leo['session'].upper()
    mission_tag = leo['mission'].upper()
    tstart =  leo['tstart'].strftime("%Y%m%dT%H%M%S")
    tend =    leo['tend'].strftime("%Y%m%dT%H%M%S")
    mydate = leo['tend']
    year = mydate.strftime('%Y') ; doy = mydate.strftime('%j')
    yy =  mydate.strftime('%y')  
    HM = mydate.strftime('%H%M') 
    datapool = os.getenv("D"); BERN_P = os.getenv("P")
    umd_data_path  = os.getenv("umd_data") 
    leo_data_path  = os.path.join(umd_data_path,leo['mission'])
    leo_att_path = f'{leo_data_path}/leoAtt/{year}.{doy}'
    subdir = kwargs.get('datapool_subdir', mission_tag) 
    attpool_path  = f'{datapool}/{subdir}/{year}{doy}'
    os.makedirs(attpool_path, exist_ok = True) 
 
    for satidx in leosats:
        leoinfo = sat_ant_info(mission, satidx=satidx)
        satname, satsvn= leoinfo['satname'], leoinfo['satsvn']
 
        # link leo files     
        fatt_file = f"{mission}_{tstart}_{tend}_{satsvn}.att"
        fatt_link = f'{satname}{year[2:]}{doy}{session_tag}_{HM}.ATT'
        att_file0 = os.path.join(leo_att_path, fatt_file)
        att_file = os.path.join(leo_att_path, fatt_link)

        flink = os.path.join(attpool_path, fatt_link)
        if os.path.isfile(att_file0):
           #flink = os.path.join(attpool_path, fatt_link)
           #shutil.move(att_file0,att_file)
           print(f'make link: {att_file0} {flink}')
           if remove_link_if_exists(flink): os.symlink(att_file0, flink)
           return True
              
        else:
           print(f'make link error: {att_file0} not exist ..')
           return False

if __name__ == "__main__":

   # Parse command-line arguments
   parser = argparse.ArgumentParser(description="Python equivalent of s06_run_bernese.bash")
   parser.add_argument("-m", "--codcom",   default="RTM", help="GNSS product")
   parser.add_argument("-t", "--session", choices=["daily", "hourly"], default="daily", help="Session type")
   parser.add_argument("--gnss-prod", choices=["COD0OPSULT","COD0OPSRAP" ],  default="COD0OPSULT", help="GNSS product")
   parser.add_argument("--gnss-latency", default=3,type=int, help="gnss productrelease latency")
   parser.add_argument("--gnss-source", choices=["NASA","AIUB"],  default="NASA", help="GNSS data source")
   parser.add_argument("--satidx", type=int, default=-1, help="Satellite index")
   parser.add_argument("--antidx", type=int, default=-1, help="Antenna index")
   parser.add_argument("mission", choices=["cosmic2", "spire"], help="Mission name")
   #parser.add_argument("sdate", help="processing date in yyyyddd or yyyymndd format")
   parser.add_argument("tstart", help="leo start time 'yyyymnddThhmmss'")
   parser.add_argument("tend", help="leo end time 'yyyymnddThhmmss'")
   args = parser.parse_args()

   stime = datetime.strptime(args.tstart, "%Y%m%dT%H%M%S")
   etime = datetime.strptime(args.tend, "%Y%m%dT%H%M%S")
   mytime = etime
   
   leo = {
     'mission': args.mission,
     'satidx':  args.satidx , 'antidx': args.antidx,
     'session': '0',
     'tstart': stime,
     'tend': tetime,
   }
   bern = {
     'codmgex': args.codcom,
     'session': args.session,
   }

   ## Date processing
   gnss_prod = 'COD0OPSULT'
   gnss_latency =  args.gnss_latency

   bernese_gnss_inputs(gnss_prod, mytime, data_source='nasa', gnss_latency = gnss_latency )
   bernese_leo_inputs_hourly(leo)



#!/bin/env python 
import argparse
import os,sys
import subprocess
from datetime import datetime, timedelta
from aiub_gnss_data import AIUBData, wget_aiub_gnss
from nasa_gnss_data import NASAData,wget_nasa_gnss

'''
#####################################################################
#
# rt_wget_gnss_ultra.py
#
# Purpose:
# This Python driver script retrieves COD0OPSULT GNSS product data
# from NASA servers for use in downstream processing.
# https://cddis.nasa.gov/archive/gnss/products/ 
#
# History:
#       Created by: Ming Chen 2026/04/24
#                   ming.chen@noaa.gov
#
# Usage: python  rt_wget_gnss_ultra.py  Args
#
# Args: datetime_str in "%Y%m%dT%H" 
#
#####################################################################

'''


def parse_args():

    # Set up main parser
    parser = argparse.ArgumentParser(description="batch driver of the leopod generation")
    parser.add_argument("-v","--verbose", action='store_true',  help="print info for debug")
    parser.add_argument("datetime", type=str, help = "%Y%m%dT%H")
 
    return parser.parse_args()

def main(): 
    args = parse_args()
    ddt = datetime.strptime(args.datetime, "%Y%m%dT%H")
    product = "COD0OPSULT"
    pdate = ddt.strftime("%Y%m%d")
    hh = ddt.strftime("%H")
    hour = int(hh)
    if hh not in ['00','06', '12','18']: sys.exit(1)
    if not NASAData.valid_product(product): sys.exit(1)
    wget_nasa_gnss(product, pdate, hour=hour)
  

if __name__ == "__main__":
    main()




#!/bin/env python 
import argparse
import os,sys, shutil
from ucar2umd_rnx_rt import umdrepo_rnx_filelist, extract_pdate_from_umdrnx_filepath
from datetime import datetime, timedelta, timezone
import re, string
from utility import  leofile_time_range, leofile_time_range_from_data
import time
from pathlib import Path

__authors__ = ["Ming Chen"]
__license__ = "Public Domain"
__maintainer__ = "Ming Chen"
__email__ = "ming.chen@noaa.gov"
__date__  = "May 08, 2026"

'''
#####################################################################
# 
# rt_file_monitor.py
#
# Purpose:
# This Python script tracks the latest available COSMIC-2 attitude
# and RINEX files to support real-time processing workflows.
#
# History:  
#       Created by: Ming Chen 2026/05/08
#                   ming.chen@noaa.gov
# 
#
#####################################################################

'''


def latest_satant_dict(mission, satidx, antidx, now_utc, **kwargs):
   
    umd_repo = kwargs.get('umd_repo','')
    #print(umd_repo)

    grace_hours = 3
    dt = timedelta(hours=grace_hours)
    
    sdate = (now_utc - dt).strftime('%Y%m%d') ; sdate = datetime.strptime(sdate,'%Y%m%d')
    edate = (now_utc).strftime('%Y%m%d') ; edate = datetime.strptime(edate,'%Y%m%d')

    filelist =[]
    while sdate <= edate:
       #print(mission, satidx, antidx, sdate, umd_repo)
       flist = umdrepo_rnx_filelist(mission, satidx, antidx, sdate, umd_repo = umd_repo)
       filelist += flist
       sdate = sdate + timedelta(days=1)  

    if len(filelist) < 1: return None

    flatest = sorted(filelist)[-1]
    flatest2 = sorted(filelist)[-2]
    #flatest = sorted(filelist,key=os.path.getmtime)[-1]
    #flatest2 = sorted(filelist,key=os.path.getmtime)[-2]
    rr=leofile_time_range(flatest)
    if  rr[1] != None :
        reftime =  datetime.strptime(rr[1],'%Y%m%dT%H:%M:%S.%f')
    fpattern = r'(\d{4}\.\d{3}\.\d{2}\.\d{2})\.(\d{3})\.(\d{2})'
    m = re.search(fpattern, flatest)
    latest_time, sat, ant = m.groups()
    current_time = datetime.strptime(latest_time, "%Y.%j.%H.%M")
    tmd = (current_time - reftime).total_seconds()
    m_time = os.path.getmtime(flatest) ; now = time.time()
    seconds_since_mod = now - m_time

    if seconds_since_mod < 60:
       print(f'{flatest}: short file lifetime  {seconds_since_mod} seconds....')
       print(current_time,reftime,tmd)
       m = re.search(fpattern, flatest2)
       latest_time, sat, ant = m.groups()
       current_time = datetime.strptime(latest_time, "%Y.%j.%H.%M")
 
    #print(current_time, sat, ant)
    return current_time


def parse_satant_file(filepath):
    result = {}
    # single regex for the whole line
    pattern = re.compile(
        r'sat:\s*(\d+)\s+ant:\s*(\d+)\s+start_time:\s*"([^"]*)"\s+end_time:\s*"([^"]*)"\s+([01])'
    )

    with open(filepath, "r") as f:
        for line in f:
            match = pattern.search(line)
            if match:
                sat, ant, start_time, end_time, status = match.groups()
                key = f"{sat}_{ant}"

                result[key] = {
                    "sat": sat,
                    "ant": ant,
                    "start_time": start_time,
                    "end_time": end_time,
                    "status": status
                }

    return result



def dump_satant_dict(data_dict, filepath):
    with open(filepath, "w") as f:
        for key, entry in data_dict.items():
            line = (
                f'sat:{entry["sat"]} '
                f'ant:{entry["ant"]} '
                f'start_time: "{entry["start_time"]}" '
                f'end_time: "{entry["end_time"]}" '
                f'{entry["status"]}\n'
            )
            f.write(line)

def update_rt_satant_dict(mission, satant_list_file,   **kwargs):

    umd_repo_default = "/mnt/uvol0/xinjia.zhou/data_GPSRO/cosmic2_nccf/level1a/podCrx"
    umd_repo = kwargs.get('umd_repo',umd_repo_default)
    now_utc = datetime.now(timezone.utc)
    ctime = now_utc 
    default_sliding_minutes = 60
    
    if not os.path.exists(satant_list_file):
       shutil.copy('cosmic2_satant.template', satant_list_file)

    satant_dict = parse_satant_file(satant_list_file)
    #print( satant_dict)
    satant_dict_new = {}
 
    sats = [1, 2, 4, 5, 6]; ants = [1, 2]
    for satidx in sats:
       for antidx in ants:
         current_time = latest_satant_dict(mission, satidx, antidx, ctime, umd_repo=umd_repo)
   
         if current_time is None: continue
     
         latest_time = current_time.strftime("%Y%m%dT%H:%M:00")
         default_start_time = current_time - timedelta(minutes=default_sliding_minutes)
         default_start_time = default_start_time.strftime("%Y%m%dT%H:%M:00")
         mykey = f'{satidx:02d}_{antidx:02d}'

         satant_dict_new[mykey] = {}
         satant_dict_new[mykey]['sat'] = f'{satidx:02d}'
         satant_dict_new[mykey]['ant'] = f'{antidx:02d}'

         if mykey in satant_dict:
            if satant_dict[mykey]['end_time'] != latest_time:
               satant_dict_new[mykey]['start_time'] = satant_dict[mykey]['end_time'] \
	       if satant_dict[mykey]['end_time'] else  default_start_time
               satant_dict_new[mykey]['end_time'] = latest_time
               satant_dict_new[mykey]['status'] = 1
            else :
               satant_dict_new[mykey]['start_time'] = satant_dict[mykey]['start_time']
               satant_dict_new[mykey]['end_time'] = satant_dict[mykey]['end_time']
               satant_dict_new[mykey]['status'] = 0

         else:
            satant_dict_new[mykey]['start_time'] = default_start_time
            satant_dict_new[mykey]['end_time'] = latest_time
            satant_dict_new[mykey]['status'] = 1
    dump_satant_dict(satant_dict_new, satant_list_file)

    return satant_dict_new
 

def get_previous_filelist(filelist_fname):
   filelist = []
   with open(filelist_fname, 'r') as file:
      for line in file:
        # Strip whitespace/newlines and skip empty lines
        file_path_str = line.strip()
        if not file_path_str:
            continue
            
        # Convert to a Path object and get the parent directory
        file_path = Path(file_path_str)
        parent_directory = file_path.parent
        basename = file_path.name
        filelist.append(basename)

   return filelist

def get_current_filelist(mission, satidx, antidx, now_utc, **kwargs):
   
    umd_repo = kwargs.get('umd_repo','')
    pdate = kwargs.get('pdate',None)
    #print(umd_repo)
    
    sdate = (now_utc).strftime('%Y%m%d')  ; sdate = datetime.strptime(sdate,'%Y%m%d')
   
    flist = umdrepo_rnx_filelist(mission, satidx, antidx, sdate, umd_repo = umd_repo)

    if pdate != None:
       pflist = umdrepo_rnx_filelist(mission, satidx, antidx, pdate, umd_repo = umd_repo)
       filelist = pflist + flist
    else:
       filelist = flist
 
    if len(filelist) < 1: return None

    return filelist


def dump_current_filelist(flist, filepath):
    with open(filepath, "w") as f:
        for line in flist:
          f.write(line+"\n")


def update_rt_filelist(mission, satant_list_file,   **kwargs):

    umd_repo_default = "/mnt/uvol0/xinjia.zhou/data_GPSRO/cosmic2_nccf/level1a/podCrx"
    umd_repo = kwargs.get('umd_repo',umd_repo_default)
    now_utc = datetime.now(timezone.utc)
    ctime = now_utc 
    N_CPU =  12
    grace_hours = 3
    dt = timedelta(hours=grace_hours)    
    midnight = now_utc.replace(hour=0, minute=0, second=0, microsecond=0)
    minutes_today = (now_utc-midnight).total_seconds() / 60
    pdate =  None
    if minutes_today < grace_hours*60:
       pdate = (now_utc - dt).strftime('%Y%m%d') 
       pdate = datetime.strptime(pdate,'%Y%m%d')
    
    if not os.path.exists(satant_list_file):
       Path(satant_list_file).touch()

    previous_list = get_previous_filelist(satant_list_file)

    sats = [1, 2, 4, 5, 6]; ants = [1, 2]
    current_list = []
    for satidx in sats:
       for antidx in ants:
         flist = get_current_filelist(mission, satidx, antidx, ctime, umd_repo=umd_repo, pdate=pdate)

         if flist == None: continue
         flist = [ Path(f).name for f in flist ] 
         current_list.extend(flist)
    

    latest =  set(current_list) - set(previous_list)
    latest = sorted(latest, reverse=True)
    latest_flist = []
    n = len(latest)
    n = n if n < N_CPU else N_CPU
     
    current_list = sorted(current_list, reverse=True)
    latest = latest[0:n]
    for flatest in latest:
       fpattern = r'(\d{4}\.\d{3}\.\d{2}\.\d{2})\.(\d{3})\.(\d{2})'
       m = re.search(fpattern, flatest)
       latest_time, sat, ant = m.groups()
       current_time = datetime.strptime(latest_time, "%Y.%j.%H.%M")
       timetag = current_time.strftime('%Y%m%dT%H:%M:%S')
       satidx=int(sat) ;  antidx = int(ant)
       latest_flist.append(f"{satidx:02d}_{antidx:02d}={timetag}")
 
    dump_current_filelist(current_list, satant_list_file)
   
    return latest_flist
  
def main():
    mission = 'cosmic2'
    satant_list_file='rt_file.list'

    ctime = datetime.strptime("2026.194.19.00", "%Y.%j.%H.%M")
    umd_repo ='/data2/mchen36/POD_DATA/umd/cosmic2/podRnx'
    latest_flist=update_rt_filelist(mission, satant_list_file, umd_repo =umd_repo)
    print(latest_flist)
    satant_list_file='rt_run_satant.list'
    latest_dict = update_rt_satant_dict(mission, satant_list_file,umd_repo =umd_repo)
    print(latest_dict)
if __name__ == "__main__":
    main()




#!/usr/bin/env python

from datetime import datetime,  timezone
import re, os


def custom_scientific_notation(x, precision=6):
    if x == 0:
        return f"{0:.{precision}f}E+00"

    import math

    sign = "-" if x < 0 else " "
    x_abs = abs(x)

    # Compute exponent so that coefficient < 1
    exponent = math.floor(math.log10(x_abs)) + 1
    coefficient = x_abs / (10 ** exponent)

    # Format the coefficient with the specified precision
    coefficient_str = f"{sign}{coefficient:.{precision}f}"

    return f"{coefficient_str}E{exponent:03d}"

# === Helper Functions ===
def parse_epoch_line(line):
    """Extract datetime from epoch header line starting with *"""
    parts = line.strip().split()
    year, month, day, hour, minute = map(int, parts[1:6])
    second = float(parts[6])
    return datetime(year, month, day, hour, minute, int(second)), second

def parse_sat_line(line):
    """Parse a satellite line starting with P and return PRN and clock bias"""
    prn = line[1:4]
    #clock_bias_str = line[-13:].strip()
    clock_bias_str = line.strip().split()[4]
    try:
        clock_bias = float(clock_bias_str)
    except ValueError:
        clock_bias = None  # Invalid entry like 999999.999999
    return prn, clock_bias


# === Main Parsing  ===

def read_records(input_file):
   epochs = []
   clock_records = {}  # {prn: [(datetime, sec_of_day, bias), ...]}

   with open(input_file, "r") as f:
      current_epoch = None
      current_sec = None
      for line in f:
        if line.startswith('*'):
            current_epoch, current_sec = parse_epoch_line(line)
            epochs.append(current_epoch)
        elif line.startswith('P'):
            prn, clock_bias = parse_sat_line(line)
            if clock_bias is not None:
                if prn not in clock_records:
                    clock_records[prn] = []
                clock_records[prn].append((current_epoch, current_sec, clock_bias*1.0e-6))
   return(clock_records)



def code_headers(dd, sats, n_sats):
   #dd = '20250629T00:00:00'
   d = datetime.strptime(dd,'%Y%m%dT%H:%M:%S')
   date_str = d.strftime("%d-%b-%y %H:%M").upper()
   prns_sorted = sorted(sats)
   code_header_lines = [
        "3.04                 C                    M                      RINEX VERSION / TYPE",
        f"CCRNXC V5.5          AIUB                 {date_str}        PGM / RUN BY / DATE ",
        "CODE RAPID PRECISE CLOCK INFORMATION                             COMMENT             ",
        "BASED ON CODE AND PHASE MEASUREMENTS                             COMMENT             ",
        "MERGED CLOCK INFORMATION FROM 5-MINUTE AND                       COMMENT             ",
        "30-SECOND CLOCK CORRECTIONS                                      COMMENT             ",
        "   GPS                                                           TIME SYSTEM ID      ",
        "    18                                                           LEAP SECONDS GNSS   ",
        "E  GPSEST V5.5        IGS20                                      SYS / PCVS APPLIED  ",
        "G  GPSEST V5.5        IGS20                                      SYS / PCVS APPLIED  ",
        "R  GPSEST V5.5        IGS20                                      SYS / PCVS APPLIED  ",
        "E  GPSEST V5.5        CODE.BIA @ ftp.aiub.unibe.ch/CODE/         SYS / DCBS APPLIED  ",
        "G  GPSEST V5.5        CODE.BIA @ ftp.aiub.unibe.ch/CODE/         SYS / DCBS APPLIED  ",
        "R  GPSEST V5.5        CODE.BIA @ ftp.aiub.unibe.ch/CODE/         SYS / DCBS APPLIED  ",
        "     2    AR    AS                                               # / TYPES OF DATA   ",
        "COD  Center for Orbit Determination in Europe                    ANALYSIS CENTER     ",
        "     1                                                           # OF CLK REF        ",
 
        "BREW00USA 40473M001                           0.000000000000E+00 ANALYSIS CLK REF    ",   
        "     1    IGS20                                                  # OF SOLN STA / TRF    ",
        "BREW00USA 40473M001           -2112007408 -3705351826  4726827014SOLN STA NAME / NUM ", 

       f"{n_sats:6d}                                                           # OF SOLN SATS"
    ]

   header_lines = code_header_lines
   # Add PRN list lines (exactly 15 IDs per line, pad with "00" if needed)
   prn_lines = []

   nid=16 ; padstr="  " ; arbias=0.0
   for i in range(0, len(prns_sorted), nid):
      chunk = prns_sorted[i:i+nid]
      # Pad chunk if less than nid
      if len(chunk) < nid:
        chunk += [padstr] * (nid - len(chunk))
      # Format line with right-aligned 3-char wide fields
      prn_line = ' '.join(f"{prn:>3}" for prn in chunk) + "  PRN LIST"
      prn_lines.append(prn_line)
   header_lines.extend(prn_lines)
   # Final COMMENT line
   header_lines.append("                                                                 END OF HEADER")
   return(header_lines)


# === Write Output File ===
def write_clk_codefmt(output_file, header_lines, clock_records):
  # Extract the directory path
  file_path = os.path.dirname(output_file)

  # Create the directory if it doesn't exist
  if file_path:  # Check if directory is not an empty string (e.g., if file_path is just a filename)
       os.makedirs(file_path, exist_ok=True)

  with open(output_file, "w") as out:
    # Write header
    for line in header_lines:
        out.write(line.rstrip() + "\n")

    # Write clock bias records
    prns_sorted = sorted(clock_records.keys())
    nrecords = len(clock_records[prns_sorted[0]])
    for i in range(nrecords):
      for p in range(len(prns_sorted)):
            prn = prns_sorted[p]
            dt, sec_of_day, bias = clock_records[prn][i]
            clock_bias = custom_scientific_notation(bias, precision=12)
            arbias = custom_scientific_notation(0.0, precision=12)
            #out.write(f"AS {prn}       {dt.year:4d} {dt.month:02d} {dt.day:02d} {dt.hour:02d} {dt.minute:02d} {dt.second:010.7f}  1 {bias:21.12E}\n")
            #out.write(f"AS {prn}       {dt.year:4d} {dt.month:02d} {dt.day:02d} {dt.hour:02d} {dt.minute:02d} {dt.second:9.6f}  2   {clock_bias}  {arbias}\n")
            out.write(f"AS {prn}       {dt.year:4d} {dt.month:02d} {dt.day:02d} {dt.hour:02d} {dt.minute:02d} {dt.second:9.6f}  1   {clock_bias}\n")

def extract_clk(input_file, output_file, datestr):

   dd = datestr
   if not os.path.exists(input_file):
      exit
   clock_records = read_records(input_file)
   n_sats = len(clock_records)
   sats = sorted(clock_records.keys())

   header_lines = code_headers(dd, sats, n_sats)
   write_clk_codefmt(output_file, header_lines, clock_records)
   print(f" RINEX clock bias file written to: {output_file}")

if __name__ == "__main__":
   import argparse

   parser = argparse.ArgumentParser(description='Main program')
   parser.add_argument("input", type=str,  help="input file")
   parser.add_argument("output", type=str,  help="output file")
   parser.add_argument("datestr", type=str,  help="date string in YYYYMMDDT00:00:00")
   parser.add_argument("-v","--verbose", action='store_true',  help="print info for debug")
 

   args = parser.parse_args()
   input_file= args.input ; output_file = args.output
   datestr = args.datestr
   if not os.path.exists(input_file):
      exit
   extract_clk(input_file, output_file, datestr)
